(* OcamlHTML - A library for generating HTML page and build CGI scripts
 *
 * html.mli - 10/9/2000
 * HTML easy generation interface
 *
 * Warning: this library uses optional arguments to make life easier
 *
 * Copyright (C) 2000 Antoine Mine' (mine@di.ens.fr)
 *)


(** HTML output functions.
   
   These functions allow you to easily output HTML page, including titles,
   forms, image, anchors, tabulars, etc.

   A subset of HTML 4.01 is implemented (see [http://www.w3.org/TR/html4/]).
   Many constructs that are not supported by older browser (Netscape 4 for
   instance) are not implemented, and some depreciated constructed still
   useful are here. Thus the produced HTML should be correct and viewable
   even to older browsers.
*)



(** {2 Utility functions} *)

val escape_URL_component: string -> string
(** [escape_URL_component s] returns a string representing [s] with special 
   characters escaped (using the [%xx] notation) to form a valid URL component.
   Beware that a complete URL may have several components separated by special
   characters [:], [/], [\@], [=], [?]; so you should not call 
   [escape_URLC_component] on a complete URL, but on each component separately!
   According to RFC 2396, URL special characters are:
   [;], [/], [?], [:], [\@], [&], [=], [+], [$], [<], [>], [#], [%], ["],
   [\{], [\}], [|], [\ ], [^], [\[], [\]], [`], [~],
   space, control characters, and non-US ASCII characters.
   An URL-escaped string can safely be inserted in HTML text without beeing
   HTML-escaped (as [<], [>], [&], and ["] are URL special characters).
*)

val escape_HTML: string -> string
(** [escape_HTML s] returns a string representing [s] with special characters
   escaped (using entities) to form a valid HTML string.
   As a side-effect, all HTML tags [<miaou>] and entities [&coincoin;] are 
   destroyed.
*)

   
(** {2 Low-level HTML generation} *)


(** All HTML generation should occur between a unique call to 
   {!Html.html_begin} and a unique call to {!Html.html_end}.
   If the HTML page is generated as a response to a HTTP request
   (HTTP server or CGI-script), do not forget to emit HTTP headers
   using {!Http.http_header} before emiting the HTML page itself.

   String arguments of the following functions can be divided in several 
   categories with respect to escape treatement:
   - liberal text between two HTML tags [<miaou>]{i text}[</miaou>] are
   not HTML-escaped (this includes paragraph contents, custom HTML);
   - as an exception to the preceding rule, code-like text within 
   [<pre>]{i text}[</pre>] is HTML-escaped;
   - attributes within a HTML tag [<miaou toto="]{i text}[">] 
   are HTML-escaped
   (this includes style class, anchor names, form field labels);
   - URL attributes ([href="]{i text}["]) are not URL-escaped, you should call
   {!Html.escape_URL_component} yourself on each URL component by hand if 
   necessary.

   These functions share many optional arguments that have the following
   meaning:
   - [?out:out_channel] output channel ([stdout] by default);
   - [?align:]{!Html.align} specify an alignement;
   - [?style:string]  specify a style class in a style-sheet (HTML-escaped);
   - [?ref:string] puts an anchor name at the current location (HTML-escaped).
*)

val html_begin: 
    ?out:out_channel            (* destination *)
 -> ?lang:string                (* two-letter lang code string *)
 -> ?author:string              (* author meta tag *)
 -> ?desc:string                (* description meta tag *)
 -> ?keywords:string list       (* keywords list for keyword meta tag *)
 -> ?styles:(string*string) list (* stylesheet (title,URL) list *)
 -> ?alternate:(string*string*string) list
                                (* alternate documents (title,lang,URL) list *)
 -> ?refresh:int                (* refresh time for refresh meta tag *)
 -> ?refresh_url:string         (* refresh URL for refresh meta tag *)
 -> ?custom:string            (* string to include unescaped in HTML header *)
 -> string                      (* title string (not escaped) *)
 -> unit
(** [html_begin title] emits a HTML header, from [<!DOCTYPE...>] to the
   opening [<body>], including the [<head>] section and many optional 
   meta-informations:
   - [title]: page title;
   - [?lang]: two-letter language code ([<html lang="...">]);
   - [?author]: the author of this page ([author] meta-tag);
   - [?desc]: description of the document ([description] meta-tag);
   - [?keywords]: a keyword list ([keywords] meta-tag);
   - [?styles]: a list of style-sheet links couples [(title,URL)];
   - [?alternate]: a list of alternate HTML page triples [(title,langugage,URL)];
   - [?refresh]: automatically refresh after [refresh] second;
   - [?refresh_url]: specify a different HTML page to load when refresh time
   [?refresh] expires;
   - [?custom]: custom text to put at the end of the [<head>] section; 
   this is not HTML-escaped.
*)

val html_end:
    ?out:out_channel   
 -> ?author:string     
 -> ?email:string      
 -> ?custom:string     (* string to include unescaped just before the end *)
 -> unit -> unit
(**
   [html_end ()] emits the a HTML footer, including an [<address>] section
   and the closing [</body>]:
   - [?author]: author name to put in the [<address>] section;
   - [?email]: email to put in the [<address>] section;
   - [?custom]: custom text (not HTML-escaped) to put at the end of the 
   [<address>] section. 
*)

    
(** Even if depreciated by Style-Sheet, it is useful to specify an
   alignement within HTML tags using this type. *)
type align = 
    Left 
  | Right 
  | Center 
  | Middle 
  | Top


val h1: 
    ?out:out_channel   
 -> ?align:align
 -> ?style:string       
 -> ?ref:string         (* anchor definition (HTML-escaped) *)
 -> string              (* title, not escaped *)
 -> unit
(** [h1 title] emits a level-1 header title [<h1>title</h1>], with an optional alignement, styleclass and anchor name. *)

val h2: 
    ?out:out_channel   
 -> ?align:align
 -> ?style:string      
 -> ?ref:string         (* anchor definition *)
 -> string -> unit
(** [h2 title] emits a level-2 header title [<h2>title</h2>], with an optional alignement, styleclass and anchor name. *)

val h3: 
    ?out:out_channel   
 -> ?align:align
 -> ?style:string      
 -> ?ref:string         (* anchor definition *)
 -> string -> unit
(** [h3 title] emits a level-3 header title [<h3>title</h3>], with an optional alignement, styleclass and anchor name. *)

val h4: 
    ?out:out_channel   
 -> ?align:align
 -> ?style:string
 -> ?ref:string         (* anchor definition *)
 -> string -> unit
(** [h4 title] emits a level-4 header title [<h4>title</h4>], with an optional alignement, styleclass and anchor name. *)


val br: ?out:out_channel -> unit -> unit
(** Emits a line break [<br>]. *)
    
val hr: ?out:out_channel -> unit -> unit
(** Emits an horizontal rule [<hr>]. *)
  
val pre: ?out:out_channel -> ?style:string -> string -> unit
(** [pre text] prints [text] in code style.
   [text] is HTML-escaped (so, do not use HTML entities like [&miaou;]) *)
   
val p: ?out:out_channel -> ?style:string -> string -> unit
(** [p text] prints [text] in the paragraph style. *)

val ul: ?out:out_channel -> ?numbered:bool -> ?style:string 
        -> string list -> unit
(** [ul list] emits the string list [list] as a HTML list. *)
    
val link: 
    ?style:string
 -> string         (* url *) 
 -> string         (* label *) 
 -> string
(** [link url text] returns the string [text] (not HTML-escaped) within
   [<a href="url">] and [</a>] tags. [url] is HTML-escaped, but you may need
   to URL-escape each URL component by hand 
   (using {!Html.escape_URL_component}). *)

(** {2 Image maps} *)
  
(** Image maps come in two different flavours: client-based (the page specify
   a set of areas in the image with attached URLs) or
   server-based (the browser sends back to the server the point where the 
   user clicked).   
   If you need a server-based image map within a form, see the function
   {!Html.form_image_map} in the form section instead.
*)

(** Area shape supported in client-side image maps. *)
type area = 
    Rectangle of (int*int)*(int*int)  (** Top-left and bottom-right corners. *)
  | Polygon   of (int*int) list       (** Vertice list. *)
  | Circle    of (int*int)*int        (** Center and radius. *)

val client_image_map: 
    ?out:out_channel
 -> ?size:int*int        (* size in pixels *)
 -> string               (* name of the map, should be unique *)
 -> string               (* URL of the image *)
 -> string               (* alternative text for image-less display *)
 -> ?default_area:string*string 
        (* default (URL,alt) for non-covered area (alt is a text desciption) *)
 -> (area*string*string) list   
        (* list of areas in the image (area,URL,alt) *)
 -> unit
(** [client_image_map name image alttext areas] emits a client-side image map
named [name] (this name must be unique in the HTML page). [image] is the URL
of the image, and [alttext] is an alternate text to be displayed if the
browser does not have imaging capabilities. Then follows in [areas] a list
of triplets [(area,url,attext)] defining areas of the image, together
with the [url] to go when the area is clicked, and an alternate
text [alttext] for image-less displays. You can provide an optional
default area [?default_area=(url,alttext)] corresponding to the parts of
the image not covered by the [(areas)] area list. *)

(* emit a server-side image map 
   uppon click, the brower will go to the URL base?x,y
*)
val server_image_map: 
    ?out:out_channel
 -> ?size:int*int        (* size in pixels *)
 -> string               (* base query URL *)
 -> string               (* URL of the image *)
 -> string               (* alternative text for image-less display *)
 -> unit
(** [server_image_map queryurl imageurl alttext] emits a server-side image map.
   [imageurl] is the URL containg the image, and [alttext] is an alternate 
   text to be displayed if the browser does not have imaging capabilities.
   When the user clicks on the image at the position (x,y), the browser will 
   go to the URL [queryurl?x,y].
   You can also specify the size of the image with the [?size] optional 
   argument. *)



(** {2 Forms} *)
  
(** This library supports several methods for sending informations from forms
   back to servers:
   - use Get for small argument sets and easy repost by bookmarking of the URL;
   - use Post for bigger argument sets;
   - use Multipart if you use file download form fields.

   Each component in a form has a string key.
   When the CGI-script calls {!Cgi.get_cgi_args}, it gets a list of pairs
   [(name,value)] that associate a string [value] to the key [name].
*)

type form_method = 
    Get            (** URL-encoded in the URL. *)
  | Post           (** URL-encoded throught stdin. *)
  | Multipart      (** Passed throught stdin as multipart MIME document. *)
  | Mailto         (** Sent in an email. *)

val form_begin: 
     ?out:out_channel 
  -> ?meth:form_method  (* method to use to pass arguments, 
                           must be Multipart for file upload *) 
  -> string             (* URL of the CGI script *)
  -> unit
(** Your form should be enclosed in [form_begin url] and [form_end ()], where
   [url] is the URL of the CGI-script that will receive the data.
   The default method is [Html.Get], but you can change this using the optional
   [?meth] argument. *)
    
val form_end: ?out:out_channel -> unit -> unit
    
val form_radio: 
    ?out:out_channel 
 -> ?checked:bool 
 -> ?label:string 
 -> string            (* name of the argument *)
 -> string            (* value of the argument *)
 -> unit
(** [form_radio arg val] emits a radio button.
   Several radios in a radio group share the same name argument [arg] and
   have different value [val].
   The radio button can have an optional textual label [?label].
   At most one button in the group can be checked using the [?checked] optional
   agrument.
   In your CGI-script, the data [val] associated with the [arg] key will 
   tell which button in the group is checked. *)

val form_radios:
     ?out:out_channel
  -> ?vertical:bool        (* alignment, horizontal by default *)
  -> ?checked:int          (* element checked by default *)
  -> string                (* name argument *)
  -> (string*string) list  (* (value,label) list of radio buttons *)
  -> unit
(** [form_radios arg vals] emits a group of several radio buttons sharing
   the same name [arg]. [vals] specify the buttons as a list of pairs
   [(val,label)].
   One radio button can be checked using the [?checked=i] optional argument
   (buttons are numbered from 0).
   In your CGI-script, the data [val] associated with the [arg] kay will 
   tell which button in the group is checked. *)

val form_text:
    ?out:out_channel 
 -> ?size:int        (* width in characters of the field displayed *)
 -> ?maxlength:int   (* max num of characters in the value *)
 -> ?default:string  (* default value of the argument *)
 -> string           (* name of the argument *)
 -> unit
(** [form_text arg] emits a one-line text field with optional visible width
   [?size], total length [?maxlength], and default value text [?default].
   In your CGI-script, the text will be associated to the key [arg]. *)
    
val form_password:
    ?out:out_channel 
 -> ?size:int        (* width in characters of the field displayed *)
 -> ?maxlength:int   (* max num of characters in the value *)
 -> ?default:string  (* default value of the argument *)
 -> string           (* name of the argument *)
 -> unit
(** [form_password] behaves as [form_text], except that the characters appear
   as [*] on the screen.
   However, the text is still passed in clear, so it is not really safe. *)
    
val form_file:
    ?out:out_channel 
 -> ?size:int*int     (* width in characters of the field displayed 
                         and height (>1 => multiple file selection possible) *)
 -> ?maxlength:int    (* max num of characters in the namefile *)
 -> ?default:string   (* default namefile *)
 -> string            (* name of the argument *)
 -> unit
(** [form_file arg] emits a file upload button.
   The filename length can be specified to be [?maxlength] at most, and a
   default file-name can be specified with [?default].
   The size of the visible file-list can be specified with [?size]; if the
   specified height is greater that one, multiple file selection is possible.
   Your CGI-script will get two informations: first, the filename string
   will be associated to the key [arg.filename]; then, the content of the file
   will be associated to the key [arg.contents] (if there are several files,
   you will get several values associated with the same key, remember that
   the filename always precedes the coresponding file contents).
   You must choose the Html.Multipart method when including this control 
in a form.*)

val form_submit: 
    ?out:out_channel
 -> ?label:string      (* label to appear in the button *)
 -> ?name:string       (* if set, you'll get name=label as argument *)
 -> unit -> unit
(** [form_submit ()] emits a sumbit button (with an optional textual label
   [?label] replacing Submit).
   If a [?name] is specified, your CGI-script will get [label] associated
   to the key [name].
   You can have several sumbit buttons with different labels or names. *)

val form_reset: 
    ?out:out_channel 
 -> ?label:string       (* label to appear in the button *)
 -> unit -> unit
(** [form_reset ()] emits a reset button (with an optional textual label
   [?label] replacing Reset). *)

val form_checkbox:
    ?out:out_channel 
 -> ?checked:bool     (* default state *)
 -> ?data:string      (* value of the argument, if checked; "on" by default *)
 -> ?label:string 
 -> string            (* name of the argument *)
 -> unit
(** [form_checkbox arg] emits a check-box with an optional label [?label].
   Originally, it is in the state [?checked] (un-checked by default).
   If checked, your CGI-script will get [data] associated to the key [arg] 
   (or [on] associated to [arg] if the optional argument [?data] is not 
   specified). *)

val form_textarea: 
    ?out:out_channel 
 -> ?readonly:bool    (* false by default *)
 -> ?default:string   (* default text in the control *)
 -> string            (* name of the argument *)
 -> int               (* number of rows *)
 -> int               (* number of columns *)
 -> unit
(** [form_textarea arg width height] emits a large text area of visible size
   [(width,height)]. An optional default text can be specified using [?default]
   and the text area can be read-only (it is not by default).
   In the CGI-script, the text will be associated with the key [arg]. *)

val form_hidden: 
    ?out:out_channel
 -> string             (* name of the argument *)
 -> string             (* value of the argument *)
 -> unit
(** [form_hiden arg val] emits an invisible form field.
   Nothing is displayed by the browser, but your CGI-script will get the
   data [val] associated to the keu [arg]. This is a simple way to attach
   short coockies to a web-page. *)

val form_image_map:
    ?out:out_channel 
 -> ?align:align         (* alignment for image *)
 -> ?size:int*int        (* size in pixels *)
 -> string               (* argument name *)
 -> string               (* URL of the image *)
 -> string               (* alternate text *)
 -> unit
(** [form_image_map arg image alttext] emits a server-side image submit button.
   [image] is the URL of the image, and [alttext] is the alternate text
   to be displayed if the browser is not image-aware. An optional size 
   in pixel can be specified with [?size=(width,height)].
   In the CGI-script, the coordinate of the mouse will be associated with
   keys [arg.x] and [arg.y]. *)

(** Hierachical popup menu. *)
type menu = 
    Option of (string option)*string*string*bool
      (** (short label, value, label, selected) *)
  | Menu of string*(menu list)      
      (** (label, option list) *) 

val form_menu:
    ?out:out_channel  
 -> ?multiple:bool    (* multiple selection, false by default *)
 -> ?size:int         (* number of visible rows *)
 -> string            (* argument name *)
 -> menu list         (* menu content *)
 -> unit
(** [form_menu arg menu] emits a hierarchical popup menu. This has not
   been tested due to browser failure... *)


