(* OcamlHTML - A library for generating HTML page and build CGI scripts
 *
 * http.mli - 10/9/2000
 * HTTP header generation interface
 *
 * Warning: this library uses optional arguments to make life easier
 *
 * Copyright (C) 2000 Antoine Mine' (mine@di.ens.fr)
 *)


(** Generate HTTP headers. 
   
   This implementation is based on RFC 2616 ([HTTP/1.1]).
*)

val get_http_message : int -> string
(** Returns the canonical status message string associated to a HTTP status
   number (e.g. 200 means [OK]). 
   Raises [Not_found] if the status number is unknown. *)

(** Some classical MIME types. *)
type mime_type = 
    MimeCustom of string*string  (** any custom MIME type [s1/s2] *)
  | MimeHTML                     (** [text/html] *)
  | MimeCSS                      (** [text/css] *)
  | MimeText                     (** [text/plain] *)
  | MimeGIF                      (** [image/gif] *)
  | MimePNG                      (** [image/png] *)
  | MimeJPEG                     (** [image/jpeg] *)
  | MimeBinary                   (** [application/octet-stream] *)

val string_of_mime : mime_type -> string
(** Returns the string defition of a MIME type. *)

(** HTTP cache control policies. *)
type cache_control =
    Public
  | Private
  | PrivateCustom of string
  | NoCache
  | NoCacheCustom of string
  | NoStore
  | NoTransform
  | MustRevalidate
  | ProxyRevalidate
  | MaxAge   of int (** in seconds *)
  | SMaxAge  of int (** in seconds *)
  | CacheCustom of string

val http_header :
    ?o:out_channel ->     (* destination *)
    ?nph:bool ->          (* Non-parsed header *)
    ?status:int ->        (* http status message number (200 OK) *)
    ?mime:mime_type ->    (* mime type *)
    ?charset:string ->    (* character set for content *)
    ?target:string ->     (* target window *)
    ?cookie:(string*string*string) list  ->(* cookies list (name,value,expiration date) *)
    ?cache:cache_control list -> (* cache control *)
    ?length:int ->       (* length of document (for non text content *)
    ?date:string ->      (* date of the document as a string (see date.ml) *)
    ?expires:string ->   (* expiration date of the document *)
    ?location:string ->  (* URL to go after expiration *)
    ?last_modified:string -> (* last modification date *)
    ?custom:(string * string) list ->(* custom headers: list of couple "a: b" *)
      unit -> unit
(** Emits a HTTP header.
   - the default is parsed-header mode for CGI-scripts; if you program a
   non-parsed headers CGI-script, or a HTTP server, set [?nph=true];
   - [?status] is a HTTP status number; if not specified, 200 ([OK]) will be choosen;
   - if not specified by [?mime], the MIME type will be [text/HTML];
   for binary contents, you should use the appropriate MIME type and
   specify the length of the file using [?length] (in bytes);
   - cookies are specified in [?cookie] as a list of triplets 
   [(name,value,expiration)] that associate [value] to the key [name]; 
   [expiration] is the expiration date of the coockie 
   (in string representation) and is mandatory;
   - all dates [?date], [?expires], [?last_modified] are in string form;
   - when an expiration date [?expires] is specifie, you can specify an
   URL [?location] to go after expiration;
   - you can specify any custom header ["head: content"] using a list
   of couples [(head,content)] in [?custom].

   After the call to [http_header] you can directly begin to emit a web-page
   using {!Html.html_begin}, or emit any data file.
 *)
