(* OcamlHTML - A library for generating HTML page and build CGI scripts
 *
 * sscookie.ml - 10/20/2000
 * Server side cookie, temporary file manipulation and
 * emulation of persistent connection to be used in CGI scripts.
 * It makes use of the sserver.ml program to send back cookies. 
 * Your CGI-script should call manually clean_cookies from time to time.
 *
 * Warning: this library uses optional arguments to make life easier
 *
 * Copyright (C) 2000 Antoine Mine' (mine@di.ens.fr)
 *)



(* Build a new random string 
   It is composed of n alphanumeric uppercase symbols (n*5 bits)
*)
let build_upper_string n =
  let ch = [|'A';'B';'C';'D';'E';'F';'G';'H';'I';'J';'K';'L';'M';
	     'N';'O';'P';'Q';'R';'S';'T';'U';'V';'W';'X';'Y';'Z';
             '1';'2';'3';'4';'5';'6' |]
  and s = String.create n
  in
  Random.self_init ();
  for i=0 to (n-1) do
    s.[i] <- ch.(((Random.bits ()) lsr 5) land 31)
  done;
  s


(* Build a filename for temporary storage
   All files are in the SSCOOKIE subdirectory of the system temporary directory
*)
let build_filename name =
  Filename.concat 
    (Filename.dirname (Filename.temp_file "" "")) 
    (Filename.concat "SSCOOKIE" name)


let setup_dir () =
  (* force creation of the temp directory *)
  (try Unix.mkdir (build_filename "") 0o777 with Unix.Unix_error _ -> ());

  (* put the readme file *)
  let readme = open_out_gen [Open_wronly;Open_binary;Open_creat;Open_trunc]
      0o644 (build_filename "00readme")
  in
  output_string readme "This directory contains server-side cookies.\nCookies are created when an OCaml program makes a call to the create_cookie function of the OCamlHTML library.\nThe cookie is retrieved using the CGI cookie server sserver shipped with the OCamlHTML library (using an URL returned by create_cookie).\nIt is also sserver's responsibility to clean outdated cookies.\n\nYou may safely destroy this directory.\n";
  close_out readme
 

(* Get the header list for a cookie.
   A Contents-length is automatically generated. 
*)
let get_headers name =
  let f = open_in_bin (build_filename name)
  and fin = ref false 
  and headers = ref [] in
  Unix.lockf (Unix.descr_of_in_channel f) Unix.F_TRLOCK 0;
  while not !fin do
    let s = input_line f in
    let i = ref 0
    and l = String.length s
    and head = Buffer.create 16
    and contents = Buffer.create 16 in
    let skip_blank () = 
      while !i<l && (s.[!i]=' ' || s.[!i]='\t') do incr i done;
    in
    if l=0 || s.[0]='\r' then fin:=true
    else
      begin
	skip_blank();
	while !i<l && s.[!i]!='\r' && s.[!i]!=':' do 
	  Buffer.add_char head s.[!i]; incr i
	done;
	if !i<l && s.[!i]=':'  then
	  begin 
	    incr i; skip_blank ();
	    while !i<l && s.[!i]!='\r' do 
	      Buffer.add_char contents s.[!i]; incr i 
	    done;
	  end;
	headers:=(Buffer.contents head,Buffer.contents contents)::(!headers)
      end
  done;
  let len = (in_channel_length f)-(pos_in f) in
  close_in f;
  let f = open_in_bin (build_filename name) in
  Unix.lockf (Unix.descr_of_in_channel f) Unix.F_ULOCK 0;
  close_in f;
  ("Content-length",string_of_int len)::(!headers)


(* returns a in-file positioned where the contents of the file begins,
   you should unlock the file after use *)
let get_contents name =
  let f = open_in_bin (build_filename name)
  and fin = ref false in
  Unix.lockf (Unix.descr_of_in_channel f) Unix.F_TRLOCK 0;
  while not !fin do
    let s = input_line f in
    if String.length s = 0 || s.[0]='\r' then fin:=true
  done;
  f
      
  

(* Clean the cookie directory *)
let clean_cookies () = 
  try
    let d = Unix.opendir (build_filename "") in
    (try while true do
      let name = Unix.readdir d in
      if name.[0]<>'0' && name.[0]<>'.' then 
	let rec toto = function
	    [] -> (try Unix.unlink (build_filename name) with _ -> ())
	  | ("Expires",s)::_ -> 
	      if Date.is_before (Date.from_string s) (Date.get_date())
	      then (try Unix.unlink (build_filename name) with _ -> ())
	  | _::a -> toto a
	in toto (get_headers name)
    done with End_of_file -> ());
    Unix.closedir d
  with _ -> ()




(* Create a new cookie and return an out_channel where to write its contents,
   together with an url where to get back the cookie.
   The channel is loocked, so that the server cannot access it while it is
   still beeing filled.
*)
let create_cookie
    ?(mime=Http.MimeHTML)  (* mime type *)
    ?charset               (* character set *)
    ?date             (* date of the document as a string (see date.ml) *)
    ?expires          (* expiration date of the document *)
    () =
  let rdate =
    (match date with
      None -> Date.to_string (Date.get_date()) | Some s -> s)
  and rexpires =
    (match expires with
      None -> let d = Date.get_date() in Date.add_minutes d 15; Date.to_string d
    | Some s -> s)
  and filename = build_upper_string 8
  and key = build_upper_string 120
  in

  setup_dir ();

  (* clean_cookies dir; *)

  (* create the temp file *)
  let o = open_out_gen [Open_wronly;Open_binary;Open_creat;Open_trunc]
       0o600 (* 0o644 *)  (build_filename filename)
  in
  
  (* locks *)
  Unix.lockf (Unix.descr_of_out_channel o) Unix.F_LOCK 0;

  (* headers *)
  Printf.fprintf o "Key: %s\r\n" key;
  Printf.fprintf o "Content-type: %s\r\n" (Http.string_of_mime mime);
  Printf.fprintf o "Date: %s\r\n" rdate;
  Printf.fprintf o "Expires: %s\r\n" rexpires;
  (match charset with None->()|Some s->Printf.fprintf o "Charset: %s\r\n" s);
  Printf.fprintf o "\r\n";
  flush o;
  
  (* build URL *)
  (o,"sserver?name="^filename^"&key="^key)
				      


